/*
 * DRM based mode setting test program
 * Copyright 2008 Tungsten Graphics
 *   Jakob Bornecrantz <jakob@tungstengraphics.com>
 * Copyright 2008 Intel Corporation
 *   Jesse Barnes <jesse.barnes@intel.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/*
 * This fairly simple test program dumps output in a similar format to the
 * "xrandr" tool everyone knows & loves.  It's necessarily slightly different
 * since the kernel separates outputs into encoder and connector structures,
 * each with their own unique ID.  The program also allows test testing of the
 * memory management and mode setting APIs by allowing the user to specify a
 * connector and mode to use for mode setting.  If all works as expected, a
 * blue background should be painted on the monitor attached to the specified
 * connector after the selected mode is set.
 *
 * TODO: use cairo to write the mode info on the selected output once
 *       the mode has been programmed, along with possible test patterns.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

#include "xf86drm.h"
#include "xf86drmMode.h"

#ifdef HAVE_CAIRO
#include <math.h>
#include <cairo.h>
#endif

drmModeRes *resources;
int fd, modes;

#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof((arr)[0]))

struct type_name {
    int type;
    char *name;
};

#define type_name_fn(res) \
char * res##_str(int type) {			\
    int i;						\
    for (i = 0; i < ARRAY_SIZE(res##_names); i++) { \
        if (res##_names[i].type == type)	\
            return res##_names[i].name;	\
    }						\
    return "(invalid)";				\
}

struct type_name encoder_type_names[] = {
    { DRM_MODE_ENCODER_NONE, "none" },
    { DRM_MODE_ENCODER_DAC, "DAC" },
    { DRM_MODE_ENCODER_TMDS, "TMDS" },
    { DRM_MODE_ENCODER_LVDS, "LVDS" },
    { DRM_MODE_ENCODER_TVDAC, "TVDAC" },
};

type_name_fn(encoder_type)

struct type_name connector_status_names[] = {
    { DRM_MODE_CONNECTED, "connected" },
    { DRM_MODE_DISCONNECTED, "disconnected" },
    { DRM_MODE_UNKNOWNCONNECTION, "unknown" },
};

type_name_fn(connector_status)

struct type_name connector_type_names[] = {
    { DRM_MODE_CONNECTOR_Unknown, "unknown" },
    { DRM_MODE_CONNECTOR_VGA, "VGA" },
    { DRM_MODE_CONNECTOR_DVII, "DVI-I" },
    { DRM_MODE_CONNECTOR_DVID, "DVI-D" },
    { DRM_MODE_CONNECTOR_DVIA, "DVI-A" },
    { DRM_MODE_CONNECTOR_Composite, "composite" },
    { DRM_MODE_CONNECTOR_SVIDEO, "s-video" },
    { DRM_MODE_CONNECTOR_LVDS, "LVDS" },
    { DRM_MODE_CONNECTOR_Component, "component" },
    { DRM_MODE_CONNECTOR_9PinDIN, "9-pin DIN" },
    { DRM_MODE_CONNECTOR_DisplayPort, "displayport" },
    { DRM_MODE_CONNECTOR_HDMIA, "HDMI-A" },
    { DRM_MODE_CONNECTOR_HDMIB, "HDMI-B" },
};

type_name_fn(connector_type)

void dump_encoders(void)
{
    drmModeEncoder *encoder;
    int i;

    printf("Encoders:\n");
    printf("id\tcrtc\ttype\tpossible crtcs\tpossible clones\t\n");
    for (i = 0; i < resources->count_encoders; i++) {
        encoder = drmModeGetEncoder(fd, resources->encoders[i]);

        if (!encoder) {
            fprintf(stderr, "could not get encoder %i: %s\n",
                resources->encoders[i], strerror(errno));
            continue;
        }
        printf("%d\t%d\t%s\t0x%08x\t0x%08x\n",
               encoder->encoder_id,
               encoder->crtc_id,
               encoder_type_str(encoder->encoder_type),
               encoder->possible_crtcs,
               encoder->possible_clones);
        drmModeFreeEncoder(encoder);
    }
    printf("\n");
}

void dump_mode(drmModeModeInfo *mode)
{
    printf("  %s %.02f %d %d %d %d %d %d %d %d\n",
           mode->name,
           (float)mode->vrefresh / 1000,
           mode->hdisplay,
           mode->hsync_start,
           mode->hsync_end,
           mode->htotal,
           mode->vdisplay,
           mode->vsync_start,
           mode->vsync_end,
           mode->vtotal);
}

static void
dump_props(drmModeConnector *connector)
{
    drmModePropertyPtr props;
    int i;

    for (i = 0; i < connector->count_props; i++) {
        props = drmModeGetProperty(fd, connector->props[i]);
        if (props == NULL)
        {
            printf("\tNULL property (%d)\n", i);
        }
        else
        {
            printf("\t%s, flags %d\n", props->name, props->flags);
        }
        drmModeFreeProperty(props);
    }
}

void dump_connectors(void)
{
    drmModeConnector *connector;
    int i, j;

    printf("Connectors:\n");
    printf("id\tencoder\tstatus\t\ttype\tsize (mm)\tmodes\n");
    for (i = 0; i < resources->count_connectors; i++) {
        connector = drmModeGetConnector(fd, resources->connectors[i]);

        if (!connector) {
            fprintf(stderr, "could not get connector %i: %s\n",
                resources->connectors[i], strerror(errno));
            continue;
        }

        printf("%d\t%d\t%s\t%s\t%dx%d\t\t%d\n",
               connector->connector_id,
               connector->encoder_id,
               connector_status_str(connector->connection),
               connector_type_str(connector->connector_type),
               connector->mmWidth, connector->mmHeight,
               connector->count_modes);

        if (!connector->count_modes)
            continue;

        printf("  modes:\n");
        printf("  name refresh (Hz) hdisp hss hse htot vdisp "
               "vss vse vtot)\n");
        for (j = 0; j < connector->count_modes; j++)
            dump_mode(&connector->modes[j]);

        drmModeFreeConnector(connector);

        printf("  props:\n");
        dump_props(connector);
    }
    printf("\n");
}

void dump_crtcs(void)
{
    drmModeCrtc *crtc;
    int i;

    printf("CRTCs:\n");
    printf("id\tfb\tpos\tsize\n");
    for (i = 0; i < resources->count_crtcs; i++) {
        crtc = drmModeGetCrtc(fd, resources->crtcs[i]);

        if (!crtc) {
            fprintf(stderr, "could not get crtc %i: %s\n",
                resources->crtcs[i], strerror(errno));
            continue;
        }
        printf("%d\t%d\t(%d,%d)\t(%dx%d)\n",
               crtc->crtc_id,
               crtc->buffer_id,
               crtc->x, crtc->y,
               crtc->width, crtc->height);
        dump_mode(&crtc->mode);

        drmModeFreeCrtc(crtc);
    }
    printf("\n");
}

void dump_framebuffers(void)
{
    drmModeFB *fb;
    int i;

    printf("Frame buffers:\n");
    printf("id\tsize\tpitch\n");
    for (i = 0; i < resources->count_fbs; i++) {
        fb = drmModeGetFB(fd, resources->fbs[i]);

        if (!fb) {
            fprintf(stderr, "could not get fb %i: %s\n",
                resources->fbs[i], strerror(errno));
            continue;
        }
        printf("%d\t(%dx%d)\t%d\n",
               fb->fb_id,
               fb->width, fb->height);

        drmModeFreeFB(fb);
    }
    printf("\n");
}

/*
 * Mode setting with the kernel interfaces is a bit of a chore.
 * First you have to find the connector in question and make sure the
 * requested mode is available.
 * Then you need to find the encoder attached to that connector so you
 * can bind it with a free crtc.
 */
struct connector {
    int id;
    char mode_str[64];
    drmModeModeInfo *mode;
    drmModeEncoder *encoder;
    int crtc;
};

static void
connector_find_mode(struct connector *c)
{
    drmModeConnector *connector;
    int i, j, size, ret, width, height;

    /* First, find the connector & mode */
    c->mode = NULL;
    for (i = 0; i < resources->count_connectors; i++) {
        connector = drmModeGetConnector(fd, resources->connectors[i]);

        if (!connector) {
            fprintf(stderr, "could not get connector %i: %s\n",
                resources->connectors[i], strerror(errno));
            drmModeFreeConnector(connector);
            continue;
        }

        if (!connector->count_modes) {
            drmModeFreeConnector(connector);
            continue;
        }

        if (connector->connector_id != c->id) {
            drmModeFreeConnector(connector);
            continue;
        }

        for (j = 0; j < connector->count_modes; j++) {
            c->mode = &connector->modes[j];
            if (!strcmp(c->mode->name, c->mode_str))
                break;
        }

        /* Found it, break out */
        if (c->mode)
            break;

        drmModeFreeConnector(connector);
    }

    if (!c->mode) {
        fprintf(stderr, "failed to find mode \"%s\"\n", c->mode_str);
        return;
    }

    /* Now get the encoder */
    for (i = 0; i < resources->count_encoders; i++) {
        c->encoder = drmModeGetEncoder(fd, resources->encoders[i]);

        if (!c->encoder) {
            fprintf(stderr, "could not get encoder %i: %s\n",
                resources->encoders[i], strerror(errno));
            drmModeFreeEncoder(c->encoder);
            continue;
        }

        if (c->encoder->encoder_id  == connector->encoder_id)
            break;

        drmModeFreeEncoder(c->encoder);
    }

    if (c->crtc == -1)
        c->crtc = c->encoder->crtc_id;
}

#define dump_resource(res) if (res) dump_##res()

int main(int argc, char **argv)
{
    int c;
    int encoders = 0, connectors = 0, crtcs = 0, framebuffers = 0;
    char *modules[] = { "i915", "radeon" };
    char *modeset = NULL, *mode, *connector;
    int i, connector_id, count = 0;
    struct connector con_args[2];

    encoders = connectors = crtcs = modes = framebuffers = 1;

    char *dev = "/dev/dri/controlD64";

    if (argc > 1)
    {
        dev = argv[1];
    }

    fd = open(dev, O_RDWR);

    resources = drmModeGetResources(fd);
    if (!resources) {
        fprintf(stderr, "drmModeGetResources failed: %s\n",
            strerror(errno));
        drmClose(fd);
        return 1;
    }

    dump_resource(encoders);
    dump_resource(connectors);
    dump_resource(crtcs);
    dump_resource(framebuffers);

    drmModeFreeResources(resources);

    return 0;
}
